@@ -23,19 +23,21 @@
 # ** end header
 #
 
+from __future__ import annotations
 from jsonschema import validate, ValidationError
+
 import os
 import re
+import datetime
 import caosdb as db
 import json
 import warnings
 from .utils import has_parent
 from .stores import GeneralStore, RecordStore
-from .structure_elements import (StructureElement, Directory, File, Dict, JSONFile,
-                                 DictIntegerElement, DictBooleanElement,
-                                 DictFloatElement, DictDictElement,
-                                 TextElement, DictTextElement, DictElement, DictListElement)
-from typing import Dict as Dict_t, List, Optional, Tuple, Union
+from .structure_elements import (StructureElement, Directory, File, DictElement, JSONFile,
+                                 IntegerElement, BooleanElement, FloatElement, NoneElement,
+                                 TextElement, TextElement, ListElement)
+from typing import List, Optional, Tuple, Union
 from abc import ABCMeta, abstractmethod
 from string import Template
 import yaml_header_tools
@@ -75,8 +77,29 @@ def str_to_bool(x):
     else:
         raise RuntimeError("Should be 'true' or 'false'.")
 
+# TODO: Comment on types and inheritance
+# Currently, we often check the type of StructureElements, because serveral converters assume that
+# they are called only with the appropriate class.
+# Raising an Error if the type is not sufficient (e.g. TextElement instead of DictElement) means
+# that the generic parent class StructureElement is actually NOT a valid type for the argument and
+# type hints should reflect this.
+# However, we should not narrow down the type of the arguments compared to the function definitions
+# in the parent Converter class. See
+# - https://mypy.readthedocs.io/en/stable/common_issues.html#incompatible-overrides
+# - https://stackoverflow.com/questions/56860/what-is-an-example-of-the-liskov-substitution-principle
+# - https://blog.daftcode.pl/covariance-contravariance-and-invariance-the-ultimate-python-guide-8fabc0c24278
+# Thus, the problem lies in the following design:
+# Converter instances are supposed to be used by the Crawler in a generic way (The crawler calls
+# `match` and `typecheck` etc) but the functions are not supposed to be called with generic
+# StructureElements. One direction out of this would be a refactoring that makes the crawler class
+# expose a generic function like `treat_element`, which can be called with any StructureElement and
+# the Converter decides what to do (e.g. do nothing if the type is one that it does not care
+# about).
+# https://gitlab.indiscale.com/caosdb/src/caosdb-crawler/-/issues/64
+
 
 class ConverterValidationError(Exception):
+
     """To be raised if contents of an element to be converted are invalid."""
 
     def __init__(self, msg):
@@ -176,9 +199,7 @@ def handle_value(value: Union[dict, str, list], values: GeneralStore):
     return (propvalue, collection_mode)
 
 
-def create_records(values: GeneralStore,
-                   records: RecordStore,
-                   def_records: dict):
+def create_records(values: GeneralStore, records: RecordStore, def_records: dict):
     # list of keys to identify, which variables have been set by which paths:
     # the items are tuples:
     # 0: record name
@@ -205,6 +226,10 @@ def create_records(values: GeneralStore,
             # additionally add the new record to the general store:
             values[name] = c_record
 
+            # add the "fallback" parent only for Records, not for Files:
+            if (role == "Record" and "parents" not in record):
+                c_record.add_parent(name)
+
         c_record = records[name]
 
         for key, value in record.items():
@@ -252,12 +277,6 @@ def create_records(values: GeneralStore,
                 var_replaced_parent = replace_variables(parent, values)
                 if not has_parent(c_record, var_replaced_parent):
                     c_record.add_parent(var_replaced_parent)
-        else:
-            # add the "fallback" parent only for Records, not for Files:
-            if role == "Record":
-                # if not has_parent(c_record, name):
-                if len(c_record.parents) == 0:
-                    c_record.add_parent(name)
     return keys_modified
 
 
@@ -266,14 +285,12 @@ class Converter(object, metaclass=ABCMeta):
     Converters treat StructureElements contained in the hierarchical sturcture.
     """
 
-    def __init__(self, definition: dict,
-                 name: str,
-                 converter_registry: dict):
+    def __init__(self, definition: dict, name: str, converter_registry: dict):
         self.definition = definition
         self.name = name
 
         # Used to store usage information for debugging:
-        self.metadata: Dict_t[str, set[str]] = {
+        self.metadata: dict[str, set[str]] = {
             "usage": set()
         }
 
@@ -286,9 +303,7 @@ class Converter(object, metaclass=ABCMeta):
                     converter_definition, converter_name, converter_registry))
 
     @staticmethod
-    def converter_factory(definition: dict,
-                          name: str,
-                          converter_registry: dict):
+    def converter_factory(definition: dict, name: str, converter_registry: dict):
         """creates a Converter instance of the appropriate class.
 
         The `type` key in the `definition` defines the Converter class which is being used.
@@ -365,28 +380,83 @@ class Converter(object, metaclass=ABCMeta):
 
         filtered_children = FILTER_FUNCTIONS[rule](to_be_filtered)
 
-        return filtered_children+unmatched_children
+        return filtered_children + unmatched_children
 
     @abstractmethod
     def typecheck(self, element: StructureElement):
+        """
+        Check whether the current structure element can be converted using
+        this converter.
+        """
         pass
 
+    @staticmethod
+    def _debug_matching_template(name: str, regexp: list[str], matched: list[str], result: Optional[dict]):
+        """ Template for the debugging output for the match function """
+        print("\n--------", name, "-----------")
+        for re, ma in zip(regexp, matched):
+            print("matching against:\n" + re)
+            print("matching:\n" + ma)
+            print("---------")
+        if result is None:
+            print("No match")
+        else:
+            print("Matched groups:")
+            print(result)
+        print("----------------------------------------")
+
+    @staticmethod
+    def debug_matching(kind=None):
+        def debug_matching_decorator(func):
+            """
+            decorator for the match function of Converters that implements debug for the match of
+            StructureElements
+            """
+
+            def inner(self, element: StructureElement):
+                mr = func(self, element)
+                if "debug_match" in self.definition and self.definition["debug_match"]:
+                    if kind == "name" and "match" in self.definition:
+                        self._debug_matching_template(name=self.__class__.__name__,
+                                                      regexp=[self.definition["match"]],
+                                                      matched=[element.name],
+                                                      result=mr)
+                    elif kind == "name_and_value":
+                        self._debug_matching_template(
+                            name=self.__class__.__name__,
+                            regexp=[self.definition["match"]
+                                    if "match" in self.definition else "",
+                                    self.definition["match_name"]
+                                    if "match_name" in self.definition else "",
+                                    self.definition["match_value"]
+                                    if "match_value" in self.definition else ""],
+                            matched=[element.name, element.name, str(element.value)],
+                            result=mr)
+                    else:
+                        self._debug_matching_template(name=self.__class__.__name__,
+                                                      regexp=self.definition["match"]
+                                                      if "match" in self.definition else "",
+                                                      matched=str(element),
+                                                      result=mr)
+                return mr
+            return inner
+        return debug_matching_decorator
+
     @abstractmethod
     def match(self, element: StructureElement) -> Optional[dict]:
+        """
+        This method is used to implement detailed checks for matching compatibility
+        of the current structure element with this converter.
+
+        The return value is a dictionary providing possible matched variables from the
+        structure elements information.
+        """
         pass
 
 
 class DirectoryConverter(Converter):
-
-    def __init__(self, definition: dict, name: str,
-                 converter_registry: dict):
-        """
-        Initialize a new directory converter.
-        """
-        super().__init__(definition, name, converter_registry)
-
-    def create_children(self, generalStore: GeneralStore,
-                        element: StructureElement):
+    def create_children(self, generalStore: GeneralStore, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, Directory):
             raise RuntimeError(
                 "Directory converters can only create children from directories.")
@@ -404,7 +474,11 @@ class DirectoryConverter(Converter):
     def typecheck(self, element: StructureElement):
         return isinstance(element, Directory)
 
+    # TODO basically all converters implement such a match function. Shouldn't this be the one
+    # of the parent class and subclasses can overwrite if needed?
+    @Converter.debug_matching("name")
     def match(self, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, Directory):
             raise RuntimeError("Element must be a directory.")
         m = re.match(self.definition["match"], element.name)
@@ -441,11 +515,12 @@ class SimpleFileConverter(Converter):
     def typecheck(self, element: StructureElement):
         return isinstance(element, File)
 
-    def create_children(self, generalStore: GeneralStore,
-                        element: StructureElement):
+    def create_children(self, generalStore: GeneralStore, element: StructureElement):
         return list()
 
+    @Converter.debug_matching("name")
     def match(self, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, File):
             raise RuntimeError("Element must be a file.")
         m = re.match(self.definition["match"], element.name)
@@ -454,16 +529,20 @@ class SimpleFileConverter(Converter):
         return m.groupdict()
 
 
-class MarkdownFileConverter(Converter):
-    def __init__(self, definition: dict, name: str,
-                 converter_registry: dict):
-        """
-        Initialize a new directory converter.
-        """
-        super().__init__(definition, name, converter_registry)
+class FileConverter(SimpleFileConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use SimpleFileConverter."))
+        super().__init__(*args, **kwargs)
 
-    def create_children(self, generalStore: GeneralStore,
-                        element: StructureElement):
+
+class MarkdownFileConverter(Converter):
+    """
+    reads the yaml header of markdown files (if a such a header exists).
+    """
+
+    def create_children(self, generalStore: GeneralStore, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, File):
             raise RuntimeError("A markdown file is needed to create children.")
 
@@ -473,9 +552,9 @@ class MarkdownFileConverter(Converter):
 
         for name, entry in header.items():
             if type(entry) == list:
-                children.append(DictListElement(name, entry))
+                children.append(ListElement(name, entry))
             elif type(entry) == str:
-                children.append(DictTextElement(name, entry))
+                children.append(TextElement(name, entry))
             else:
                 raise RuntimeError(
                     "Header entry {} has incompatible type.".format(name))
@@ -484,7 +563,9 @@ class MarkdownFileConverter(Converter):
     def typecheck(self, element: StructureElement):
         return isinstance(element, File)
 
+    @Converter.debug_matching("name")
     def match(self, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, File):
             raise RuntimeError("Element must be a file.")
         m = re.match(self.definition["match"], element.name)
@@ -499,58 +580,118 @@ class MarkdownFileConverter(Converter):
         return m.groupdict()
 
 
-class DictConverter(Converter):
-    # TODO use Dict as typecheck?
-    def create_children(self, generalStore: GeneralStore, element: StructureElement):
-        if not self.typecheck(element):
-            raise RuntimeError("A dict is needed to create children")
+def convert_basic_element(element: Union[list, dict, bool, int, float, str, None], name=None,
+                          msg_prefix=""):
+    """converts basic Python objects to the corresponding StructureElements """
+    if isinstance(element, list):
+        return ListElement(name, element)
+    elif isinstance(element, dict):
+        return DictElement(name, element)
+    elif isinstance(element, bool):
+        return BooleanElement(name, element)
+    elif isinstance(element, int):
+        return IntegerElement(name, element)
+    elif isinstance(element, float):
+        return FloatElement(name, element)
+    elif isinstance(element, str):
+        return TextElement(name, element)
+    elif element is None:
+        return NoneElement(name)
+    elif isinstance(element, datetime.date):
+        return TextElement(name, str(element))
+    else:
+        raise NotImplementedError(
+            msg_prefix + f"The object that has an unexpected type: {type(element)}\n"
+            f"The object is:\n{str(element)}")
 
-        return self._create_children_from_dict(element.value)
+
+def validate_against_json_schema(instance, schema_resource: Union[dict, str]):
+    """validates given ``instance`` against given ``schema_resource``.
+
+    Args:
+        instance: instance to be validated, typically ``dict`` but can be ``list``, ``str``, etc.
+        schema_resource: Either a path to the JSON file containing the schema or a  ``dict`` with
+        the schema
+    """
+    if isinstance(schema_resource, dict):
+        schema = schema_resource
+    elif isinstance(schema_resource, str):
+        with open(schema_resource, 'r') as json_file:
+            schema = json.load(json_file)
+    else:
+        raise ValueError("The value of 'validate' has to be a string describing the path "
+                         "to the json schema file (relative to the cfood yml)  "
+                         "or a dict containing the schema.")
+    # validate instance (e.g. JSON content) against schema
+    try:
+        validate(instance=instance, schema=schema)
+    except ValidationError as err:
+        raise ConverterValidationError(
+            f"\nCouldn't validate {instance}:\n{err.message}")
+
+
+class DictElementConverter(Converter):
+    def create_children(self, generalStore: GeneralStore, element: StructureElement):
+        # TODO: See comment on types and inheritance
+        if not isinstance(element, DictElement):
+            raise ValueError("create_children was called with wrong type of StructureElement")
+
+        try:
+            return self._create_children_from_dict(element.value)
+        except ConverterValidationError as err:
+            path = generalStore[self.name]
+            raise ConverterValidationError(
+                "Error during the validation of the dictionary located at the following node "
+                "in the data structure:\n"
+                f"{path}\n" + err.message)
 
     def _create_children_from_dict(self, data):
+        if "validate" in self.definition and self.definition["validate"]:
+            validate_against_json_schema(data, self.definition["validate"])
+
         children = []
 
         for name, value in data.items():
-            if type(value) == list:
-                children.append(DictListElement(name, value))
-            elif type(value) == str:
-                children.append(DictTextElement(name, value))
-            elif type(value) == dict:
-                children.append(DictDictElement(name, value))
-            elif type(value) == int:
-                children.append(DictIntegerElement(name, value))
-            elif type(value) == bool:
-                children.append(DictBooleanElement(name, value))
-            elif type(value) == float:
-                children.append(DictFloatElement(name, value))
-            elif type(value) == type(None):
-                continue
-            else:
-                children.append(DictElement(name, value))
-                warnings.warn(f"The value in the dict for key:{name} has an unknown type. "
-                              "The fallback type DictElement is used.")
+            children.append(convert_basic_element(
+                value, name, f"The value in the dict for key:{name} has an unknown type."))
 
         return children
 
-    # TODO use Dict as typecheck?
     def typecheck(self, element: StructureElement):
-        return isinstance(element, Dict)
+        return isinstance(element, DictElement)
 
+    @Converter.debug_matching("name_and_value")
     def match(self, element: StructureElement):
         """
         Allways matches if the element has the right type.
         """
-        if not isinstance(element, Dict):
+        # TODO: See comment on types and inheritance
+        if not isinstance(element, DictElement):
             raise RuntimeError("Element must be a DictElement.")
-        return {}
+        return match_name_and_value(self.definition, element.name, element.value)
 
 
-# TODO: difference to SimpleFileConverter? Do we need both?
-class FileConverter(Converter):
+class DictConverter(DictElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use DictConverter."))
+        super().__init__(*args, **kwargs)
+
+
+class DictDictElementConverter(DictElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use DictElementConverter."))
+        super().__init__(*args, **kwargs)
+
+
+class JSONFileConverter(Converter):
     def typecheck(self, element: StructureElement):
         return isinstance(element, File)
 
+    @Converter.debug_matching("name")
     def match(self, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not self.typecheck(element):
             raise RuntimeError("Element must be a file")
         m = re.match(self.definition["match"], element.name)
@@ -559,59 +700,132 @@ class FileConverter(Converter):
         return m.groupdict()
 
     def create_children(self, generalStore: GeneralStore, element: StructureElement):
-        return []
-
-
-class JSONFileConverter(DictConverter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, File)
-
-    def match(self, element: StructureElement):
-        if not self.typecheck(element):
-            raise RuntimeError("Element must be a file")
-        m = re.match(self.definition["match"], element.name)
-        if m is None:
-            return None
-        return m.groupdict()
-
-    def create_children(self, generalStore: GeneralStore, element: StructureElement):
-        if not self.typecheck(element):
-            raise RuntimeError("A JSON file is needed to create children")
-        # TODO: either add explicit time check for File structure element here,
-        #       or add a comment to suppress mypy type warning.
+        # TODO: See comment on types and inheritance
+        if not isinstance(element, File):
+            raise ValueError("create_children was called with wrong type of StructureElement")
         with open(element.path, 'r') as json_file:
             json_data = json.load(json_file)
-        if not isinstance(json_data, dict):
-            raise NotImplementedError("JSON file must contain a dict")
         if "validate" in self.definition and self.definition["validate"]:
-            if isinstance(self.definition["validate"], dict):
-                schema = self.definition["validate"]
-            elif isinstance(self.definition["validate"], str):
-
-                with open(self.definition["validate"], 'r') as json_file:
-                    schema = json.load(json_file)
-            else:
-                raise ValueError("The value of 'validate' has to be a string describing the path "
-                                 "to the json schema file (relative to the cfood yml)  "
-                                 "or a dict containing the schema.")
-            # Validate the json content
             try:
-                validate(instance=json_data, schema=schema)
-            except ValidationError as err:
+                validate_against_json_schema(json_data, self.definition["validate"])
+            except ConverterValidationError as err:
                 raise ConverterValidationError(
-                    f"Couldn't validate {json_data}:\n{err.message}")
-
-        return self._create_children_from_dict(json_data)
+                    "Error during the validation of the JSON file:\n"
+                    f"{element.path}\n" + err.message)
+        structure_element = convert_basic_element(
+            json_data,
+            name=element.name+"_child_dict",
+            msg_prefix="The JSON File contained content that was parsed to a Python object"
+            " with an unexpected type.")
+        return [structure_element]
 
 
-class _AbstractDictElementConverter(Converter):
+class YAMLFileConverter(Converter):
+    def typecheck(self, element: StructureElement):
+        return isinstance(element, File)
+
+    @Converter.debug_matching("name")
+    def match(self, element: StructureElement):
+        # TODO: See comment on types and inheritance
+        if not self.typecheck(element):
+            raise RuntimeError("Element must be a file")
+        m = re.match(self.definition["match"], element.name)
+        if m is None:
+            return None
+        return m.groupdict()
+
+    def create_children(self, generalStore: GeneralStore, element: StructureElement):
+        # TODO: See comment on types and inheritance
+        if not isinstance(element, File):
+            raise ValueError("create_children was called with wrong type of StructureElement")
+        with open(element.path, 'r') as yaml_file:
+            yaml_data = yaml.safe_load(yaml_file)
+        if "validate" in self.definition and self.definition["validate"]:
+            try:
+                validate_against_json_schema(yaml_data, self.definition["validate"])
+            except ConverterValidationError as err:
+                raise ConverterValidationError(
+                    "Error during the validation of the YAML file:\n"
+                    f"{element.path}\n" + err.message)
+        structure_element = convert_basic_element(
+            yaml_data,
+            name=element.name+"_child_dict",
+            msg_prefix="The YAML File contained content that was parsed to a Python object"
+            " with an unexpected type.")
+        return [structure_element]
+
+
+def match_name_and_value(definition, name, value):
+    """
+    takes match definitions from the definition argument and applies regular expressiion to name
+    and possibly value
+
+    one of the keys 'match_name' and "match' needs to be available in definition
+    'match_value' is optional
+
+    Returns None, if match_name or match lead to no match. Otherwise, returns a dictionary with the
+        matched groups, possibly including matches from using match_value
+    """
+    if "match_name" in definition:
+        if "match" in definition:
+            raise RuntimeError(f"Do not supply both, 'match_name' and 'match'.")
+
+        m1 = re.match(definition["match_name"], name)
+        if m1 is None:
+            return None
+        else:
+            m1 = m1.groupdict()
+    elif "match" in definition:
+        m1 = re.match(definition["match"], name)
+        if m1 is None:
+            return None
+        else:
+            m1 = m1.groupdict()
+    else:
+        m1 = {}
+
+    if "match_value" in definition:
+        m2 = re.match(definition["match_value"], str(value), re.DOTALL)
+        if m2 is None:
+            return None
+        else:
+            m2 = m2.groupdict()
+    else:
+        m2 = {}
+
+    values = dict()
+    values.update(m1)
+    values.update(m2)
+    return values
+
+
+class _AbstractScalarValueElementConverter(Converter):
+    """
+    A base class for all converters that have a scalar value that can be matched using a regular
+    expression.
+
+    values must have one of the following type: str, bool, int, float
+    """
+
+    default_matches = {
+        "accept_text": False,
+        "accept_bool": False,
+        "accept_int": False,
+        "accept_float": False,
+    }
 
     def create_children(self, generalStore: GeneralStore, element: StructureElement):
         return []
 
     def typecheck(self, element: StructureElement):
-        return True
+        """
+        returns whether the type of StructureElement is accepted by this converter instance.
+        """
+        allowed_matches = self._merge_match_definition_with_default(self.default_matches,
+                                                                    self.definition)
+        return self._typecheck(element, allowed_matches)
 
+    @Converter.debug_matching("name_and_value")
     def match(self, element: StructureElement):
         """
         Try to match the given structure element.
@@ -621,111 +835,167 @@ class _AbstractDictElementConverter(Converter):
         Else return a dictionary containing the variables from the matched regexp
         as key value pairs.
         """
-        if not self.typecheck(element):
-            raise RuntimeError(
-                f"Element has an invalid type: {type(element)}.")
-        m1 = re.match(self.definition["match_name"], element.name)
-        if m1 is None:
-            return None
-        m2 = re.match(self.definition["match_value"], str(element.value))
-        if m2 is None:
-            return None
-        values = dict()
-        values.update(m1.groupdict())
-        values.update(m2.groupdict())
-        return values
+        # TODO: See comment on types and inheritance
+        if (not isinstance(element, TextElement)
+                and not isinstance(element, BooleanElement)
+                and not isinstance(element, IntegerElement)
+                and not isinstance(element, FloatElement)):
+            raise ValueError("create_children was called with wrong type of StructureElement")
+        return match_name_and_value(self.definition, element.name, element.value)
+
+    def _typecheck(self, element: StructureElement, allowed_matches: dict):
+        """
+        returns whether the type of StructureElement is accepted.
+
+        Parameters:
+        element: StructureElement, the element that is checked
+        allowed_matches: Dict, a dictionary that defines what types are allowed. It must have the
+                         keys 'accept_text', 'accept_bool', 'accept_int', and 'accept_float'.
+
+        returns:  whether or not the converter allows the type of element
+        """
+        if (bool(allowed_matches["accept_text"]) and isinstance(element, TextElement)):
+            return True
+        elif (bool(allowed_matches["accept_bool"]) and isinstance(element, BooleanElement)):
+            return True
+        elif (bool(allowed_matches["accept_int"]) and isinstance(element, IntegerElement)):
+            return True
+        elif (bool(allowed_matches["accept_float"]) and isinstance(element, FloatElement)):
+            return True
+        else:
+            return False
+
+    def _merge_match_definition_with_default(self, default: dict, definition: dict):
+        """
+        returns a dict with the same keys as default dict but with updated values from definition
+        where it has the same keys
+        """
+
+        result = {}
+        for key in default:
+            if key in definition:
+                result[key] = definition[key]
+            else:
+                result[key] = default[key]
+        return result
 
 
-class DictBooleanElementConverter(_AbstractDictElementConverter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, DictBooleanElement)
+class BooleanElementConverter(_AbstractScalarValueElementConverter):
+    default_matches = {
+        "accept_text": False,
+        "accept_bool": True,
+        "accept_int": True,
+        "accept_float": False,
+    }
 
 
-class DictFloatElementConverter(_AbstractDictElementConverter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, DictFloatElement)
+class DictBooleanElementConverter(BooleanElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use BooleanElementConverter."))
+        super().__init__(*args, **kwargs)
 
 
-class DictTextElementConverter(_AbstractDictElementConverter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, DictTextElement)
+class FloatElementConverter(_AbstractScalarValueElementConverter):
+    default_matches = {
+        "accept_text": False,
+        "accept_bool": False,
+        "accept_int": True,
+        "accept_float": True,
+    }
 
 
-class DictIntegerElementConverter(_AbstractDictElementConverter):
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, DictIntegerElement)
+class DictFloatElementConverter(FloatElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use FloatElementConverter."))
+        super().__init__(*args, **kwargs)
 
 
-class DictListElementConverter(Converter):
+class TextElementConverter(_AbstractScalarValueElementConverter):
+    default_matches = {
+        "accept_text": True,
+        "accept_bool": True,
+        "accept_int": True,
+        "accept_float": True,
+    }
+
+    def __init__(self, definition, *args, **kwargs):
+        if "match" in definition:
+            raise ValueError("""
+The 'match' key will in future be used to match a potential name of a TextElement. Please use
+the 'match_value' key to match the value of the TextElement and 'match_name' for matching the name.
+""")
+
+        super().__init__(definition, *args, **kwargs)
+
+
+class DictTextElementConverter(TextElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use TextElementConverter."))
+        super().__init__(*args, **kwargs)
+
+
+class IntegerElementConverter(_AbstractScalarValueElementConverter):
+    default_matches = {
+        "accept_text": False,
+        "accept_bool": False,
+        "accept_int": True,
+        "accept_float": False,
+    }
+
+
+class DictIntegerElementConverter(IntegerElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use IntegerElementConverter."))
+        super().__init__(*args, **kwargs)
+
+
+class ListElementConverter(Converter):
     def create_children(self, generalStore: GeneralStore,
                         element: StructureElement):
-        if not isinstance(element, DictListElement):
+        # TODO: See comment on types and inheritance
+        if not isinstance(element, ListElement):
             raise RuntimeError(
                 "This converter can only process DictListElements.")
-        children = []
+        children: list[StructureElement] = []
         for index, list_element in enumerate(element.value):
             # TODO(fspreck): Refactor this and merge with DictXXXElements maybe?
             if isinstance(list_element, str):
                 children.append(TextElement(str(index), list_element))
             elif isinstance(list_element, dict):
-                children.append(Dict(str(index), list_element))
+                children.append(DictElement(str(index), list_element))
+            elif isinstance(list_element, StructureElement):
+                children.append(list_element)
             else:
                 raise NotImplementedError(
                     f"Unkown type {type(list_element)} in list element {list_element}.")
         return children
 
     def typecheck(self, element: StructureElement):
-        return isinstance(element, DictListElement)
+        return isinstance(element, ListElement)
 
+    @Converter.debug_matching("name")
     def match(self, element: StructureElement):
-        if not isinstance(element, DictListElement):
-            raise RuntimeError("Element must be a DictListElement.")
+        # TODO: See comment on types and inheritance
+        if not isinstance(element, ListElement):
+            raise RuntimeError("Element must be a ListElement.")
         m = re.match(self.definition["match_name"], element.name)
         if m is None:
             return None
         if "match" in self.definition:
             raise NotImplementedError(
-                "Match is not implemented for DictListElement.")
+                "Match is not implemented for ListElement.")
         return m.groupdict()
 
 
-class DictDictElementConverter(DictConverter):
-    def create_children(self, generalStore: GeneralStore, element: StructureElement):
-        if not self.typecheck(element):
-            raise RuntimeError("A dict is needed to create children")
-
-        return self._create_children_from_dict(element.value)
-
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, DictDictElement)
-
-    def match(self, element: StructureElement):
-        if not self.typecheck(element):
-            raise RuntimeError("Element must be a DictDictElement.")
-        m = re.match(self.definition["match_name"], element.name)
-        if m is None:
-            return None
-        if "match" in self.definition:
-            raise NotImplementedError(
-                "Match is not implemented for DictDictElement.")
-        return m.groupdict()
-
-
-class TextElementConverter(Converter):
-    def create_children(self, generalStore: GeneralStore,
-                        element: StructureElement):
-        return []
-
-    def typecheck(self, element: StructureElement):
-        return isinstance(element, TextElement)
-
-    def match(self, element: StructureElement):
-        if not isinstance(element, TextElement):
-            raise RuntimeError("Element must be a TextElement.")
-        m = re.match(self.definition["match"], element.value)
-        if m is None:
-            return None
-        return m.groupdict()
+class DictListElementConverter(ListElementConverter):
+    def __init__(self, *args, **kwargs):
+        warnings.warn(DeprecationWarning(
+            "This class is depricated. Please use ListElementConverter."))
+        super().__init__(*args, **kwargs)
 
 
 class TableConverter(Converter):
@@ -733,10 +1003,10 @@ class TableConverter(Converter):
     This converter reads tables in different formats line by line and
     allows matching the corresponding rows.
 
-    The subtree generated by the table converter consists of DictDictElements, each being
+    The subtree generated by the table converter consists of DictElements, each being
     a row. The corresponding header elements will become the dictionary keys.
 
-    The rows can be matched using a DictDictElementConverter.
+    The rows can be matched using a DictElementConverter.
     """
     @abstractmethod
     def get_options(self):
@@ -752,7 +1022,8 @@ class TableConverter(Converter):
             if opt_name in self.definition:
                 el = self.definition[opt_name]
                 # The option can often either be a single value or a list of values.
-                # In the latter case each element of the list will be converted to the defined type.
+                # In the latter case each element of the list will be converted to the defined
+                # type.
                 if isinstance(el, list):
                     option_dict[opt_name] = [
                         opt_conversion(el_el) for el_el in el]
@@ -763,7 +1034,9 @@ class TableConverter(Converter):
     def typecheck(self, element: StructureElement):
         return isinstance(element, File)
 
+    @Converter.debug_matching("name")
     def match(self, element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, File):
             raise RuntimeError("Element must be a File.")
         m = re.match(self.definition["match"], element.name)
@@ -790,13 +1063,14 @@ class XLSXTableConverter(TableConverter):
 
     def create_children(self, generalStore: GeneralStore,
                         element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, File):
             raise RuntimeError("Element must be a File.")
         table = pd.read_excel(element.path, **self.get_options())
         child_elements = list()
         for index, row in table.iterrows():
             child_elements.append(
-                DictDictElement(str(index), row.to_dict()))
+                DictElement(str(index), row.to_dict()))
         return child_elements
 
 
@@ -818,11 +1092,12 @@ class CSVTableConverter(TableConverter):
 
     def create_children(self, generalStore: GeneralStore,
                         element: StructureElement):
+        # TODO: See comment on types and inheritance
         if not isinstance(element, File):
             raise RuntimeError("Element must be a File.")
         table = pd.read_csv(element.path, **self.get_options())
         child_elements = list()
         for index, row in table.iterrows():
             child_elements.append(
-                DictDictElement(str(index), row.to_dict()))
+                DictElement(str(index), row.to_dict()))
         return child_elements
