@@ -23,9 +23,12 @@
 # ** end header
 #
 
+from __future__ import annotations
 import yaml
 
 from datetime import datetime
+from typing import Any
+from .identifiable import Identifiable
 import caosdb as db
 import logging
 from abc import abstractmethod, ABCMeta
@@ -33,14 +36,14 @@ from .utils import has_parent
 logger = logging.getLogger(__name__)
 
 
-def convert_value(value):
+def convert_value(value: Any):
     """ Returns a string representation of the value that is suitable
     to be used in the query
     looking for the identified record.
 
     Parameters
     ----------
-    value : The property of which the value shall be returned.
+    value : Any type, the value that shall be returned and potentially converted.
 
     Returns
     -------
@@ -52,11 +55,13 @@ def convert_value(value):
         return str(value.id)
     elif isinstance(value, datetime):
         return value.isoformat()
-    elif type(value) == str:
+    elif isinstance(value, bool):
+        return str(value).upper()
+    elif isinstance(value, str):
         # replace single quotes, otherwise they may break the queries
         return value.replace("\'", "\\'")
     else:
-        return f"{value}"
+        return str(value)
 
 
 class IdentifiableAdapter(metaclass=ABCMeta):
@@ -86,7 +91,7 @@ class IdentifiableAdapter(metaclass=ABCMeta):
     """
 
     @staticmethod
-    def create_query_for_identifiable(ident: db.Record):
+    def create_query_for_identifiable(ident: Identifiable):
         """
         This function is taken from the old crawler:
         caosdb-advanced-user-tools/src/caosadvancedtools/crawler.py
@@ -95,34 +100,34 @@ class IdentifiableAdapter(metaclass=ABCMeta):
         whether the required record already exists.
         """
 
-        if len(ident.parents) != 1:
-            raise RuntimeError(
-                "Multiple parents for identifiables not supported.")
+        query_string = "FIND RECORD "
+        if ident.record_type is not None:
+            query_string += ident.record_type
+        for ref in ident.backrefs:
+            eid = ref
+            if isinstance(ref, db.Entity):
+                eid = ref.id
+            query_string += (" WHICH IS REFERENCED BY " + str(eid) + " AND")
 
-        query_string = "FIND Record " + ident.get_parents()[0].name
         query_string += " WITH "
 
-        if ident.name is None and len(ident.get_properties()) == 0:
-            raise ValueError(
-                "The identifiable must have features to identify it.")
-
         if ident.name is not None:
             query_string += "name='{}'".format(ident.name)
-            if len(ident.get_properties()) > 0:
+            if len(ident.properties) > 0:
                 query_string += " AND "
 
         query_string += IdentifiableAdapter.create_property_query(ident)
         return query_string
 
     @staticmethod
-    def create_property_query(entity: db.Entity):
+    def create_property_query(entity: Identifiable):
         query_string = ""
-        for p in entity.get_properties():
-            if p.value is None:
-                query_string += "'" + p.name + "' IS NULL AND "
-            elif isinstance(p.value, list):
-                for v in p.value:
-                    query_string += ("'" + p.name + "'='" +
+        for pname, pvalue in entity.properties.items():
+            if pvalue is None:
+                query_string += "'" + pname + "' IS NULL AND "
+            elif isinstance(pvalue, list):
+                for v in pvalue:
+                    query_string += ("'" + pname + "'='" +
                                      convert_value(v) + "' AND ")
 
             # TODO: (for review)
@@ -136,8 +141,8 @@ class IdentifiableAdapter(metaclass=ABCMeta):
             #                      IdentifiableAdapter.create_property_query(p.value) +
             #                      ") AND ")
             else:
-                query_string += ("'" + p.name + "'='" +
-                                 convert_value(p.value) + "' AND ")
+                query_string += ("'" + pname + "'='" +
+                                 convert_value(pvalue) + "' AND ")
         # remove the last AND
         return query_string[:-4]
 
@@ -160,79 +165,90 @@ class IdentifiableAdapter(metaclass=ABCMeta):
         """
         pass
 
-    def get_identifiable_for_file(self, record: db.File):
-        """
-        Retrieve an identifiable for a file.
-
-        Currently an identifiable for a file ist just a File object
-        with a specific path. In the future, this could be extended
-        to allow for names, parents and custom properties.
-        """
-        identifiable = db.File()
-        identifiable.path = record.path
-        return identifiable
-
-    def get_identifiable(self, record: db.Record):
+    def get_identifiable(self, record: db.Record, referencing_entities=None):
         """
         retrieve the registred identifiable and fill the property values to create an
         identifiable
-        """
 
-        if record.role == "File":
-            return self.get_identifiable_for_file(record)
+        Args:
+            record: the record for which the Identifiable shall be created.
+            referencing_entities: a dictionary (Type: dict[int, dict[str, list[db.Entity]]]), that
+              allows to look up entities with a certain RecordType, that reference ``record``
+
+        Returns:
+            Identifiable, the identifiable for record.
+        """
 
         registered_identifiable = self.get_registered_identifiable(record)
 
-        if registered_identifiable is None:
-            return None
+        if referencing_entities is None:
+            referencing_entities = {}
 
-        identifiable = db.Record(name=record.name)
-        if len(registered_identifiable.parents) != 1:
-            raise RuntimeError("Multiple parents for identifiables"
-                               "not supported.")
-        identifiable.add_parent(registered_identifiable.parents[0])
         property_name_list_A = []
         property_name_list_B = []
+        identifiable_props = {}
+        identifiable_backrefs = []
 
-        # fill the values:
-        for prop in registered_identifiable.properties:
-            if prop.name == "name":
-                # The name can be an identifiable, but it isn't a property
-                continue
-            # problem: what happens with multi properties?
-            # case A: in the registered identifiable
-            # case B: in the identifiable
+        if registered_identifiable is not None:
+            # fill the values:
+            for prop in registered_identifiable.properties:
+                if prop.name == "name":
+                    # The name can be an identifiable, but it isn't a property
+                    continue
+                # problem: what happens with multi properties?
+                # case A: in the registered identifiable
+                # case B: in the identifiable
 
-            record_prop = record.get_property(prop.name)
-            if record_prop is None:
-                # TODO: how to handle missing values in identifiables
-                #       raise an exception?
-                raise NotImplementedError(
-                    f"The following record is missing an identifying property:"
-                    f"RECORD\n{record}\nIdentifying PROPERTY\n{prop.name}"
-                )
-            newval = record_prop.value
-            record_prop_new = db.Property(name=record_prop.name,
-                                          id=record_prop.id,
-                                          description=record_prop.description,
-                                          datatype=record_prop.datatype,
-                                          value=newval,
-                                          unit=record_prop.unit)
-            identifiable.add_property(record_prop_new)
-            property_name_list_A.append(prop.name)
+                # TODO: similar to the Identifiable class, Registred Identifiable should be a
+                # separate class too
+                if prop.name.lower() == "is_referenced_by":
+                    for rtname in prop.value:
+                        if (id(record) in referencing_entities
+                                and rtname in referencing_entities[id(record)]):
+                            identifiable_backrefs.extend(referencing_entities[id(record)][rtname])
+                        else:
+                            # TODO: is this the appropriate error?
+                            raise NotImplementedError(
+                                f"The following record is missing an identifying property:"
+                                f"RECORD\n{record}\nIdentifying PROPERTY\n{prop.name}"
+                            )
+                    continue
 
-        # check for multi properties in the record:
-        for prop in property_name_list_A:
-            property_name_list_B.append(prop)
-        if (len(set(property_name_list_B)) != len(property_name_list_B) or len(
-                set(property_name_list_A)) != len(property_name_list_A)):
-            raise RuntimeError(
-                "Multi properties used in identifiables can cause unpredictable results.")
+                record_prop = record.get_property(prop.name)
+                if record_prop is None:
+                    # TODO: how to handle missing values in identifiables
+                    #       raise an exception?
+                    # TODO: is this the appropriate error?
+                    raise NotImplementedError(
+                        f"The following record is missing an identifying property:"
+                        f"RECORD\n{record}\nIdentifying PROPERTY\n{prop.name}"
+                    )
+                identifiable_props[record_prop.name] = record_prop.value
+                property_name_list_A.append(prop.name)
 
-        return identifiable
+            # check for multi properties in the record:
+            for prop in property_name_list_A:
+                property_name_list_B.append(prop)
+            if (len(set(property_name_list_B)) != len(property_name_list_B) or len(
+                    set(property_name_list_A)) != len(property_name_list_A)):
+                raise RuntimeError(
+                    "Multi properties used in identifiables could cause unpredictable results and "
+                    "are not allowed. You might want to consider a Property with a list as value.")
+
+        # use the RecordType of the registred Identifiable if it exists
+        # We do not use parents of Record because it might have multiple
+        return Identifiable(
+            record_id=record.id,
+            record_type=(registered_identifiable.parents[0].name
+                         if registered_identifiable else None),
+            name=record.name,
+            properties=identifiable_props,
+            path=record.path,
+            backrefs=identifiable_backrefs
+        )
 
     @abstractmethod
-    def retrieve_identified_record_for_identifiable(self, identifiable: db.Record):
+    def retrieve_identified_record_for_identifiable(self, identifiable: Identifiable):
         """
         Retrieve identifiable record for a given identifiable.
 
@@ -245,7 +261,7 @@ class IdentifiableAdapter(metaclass=ABCMeta):
 
     # TODO: remove side effect
     # TODO: use ID if record has one?
-    def retrieve_identified_record_for_record(self, record: db.Record):
+    def retrieve_identified_record_for_record(self, record: db.Record, referencing_entities=None):
         """
         This function combines all functionality of the IdentifierAdapter by
         returning the identifiable after having checked for an appropriate
@@ -254,12 +270,9 @@ class IdentifiableAdapter(metaclass=ABCMeta):
         In case there was no appropriate registered identifiable or no identifiable could
         be found return value is None.
         """
-        identifiable = self.get_identifiable(record)
+        identifiable = self.get_identifiable(record, referencing_entities=referencing_entities)
 
-        if identifiable is None:
-            return None
-
-        if identifiable.role == "File":
+        if identifiable.path is not None:
             return self.get_file(identifiable)
 
         return self.retrieve_identified_record_for_identifiable(identifiable)
@@ -280,7 +293,7 @@ class LocalStorageIdentifiableAdapter(IdentifiableAdapter):
     def get_records(self):
         return self._records
 
-    def get_file(self, identifiable: db.File):
+    def get_file(self, identifiable: Identifiable):
         """
         Just look in records for a file with the same path.
         """
@@ -338,7 +351,7 @@ class LocalStorageIdentifiableAdapter(IdentifiableAdapter):
             return None
         return identifiable_candidates[0]
 
-    def check_record(self, record: db.Record, identifiable: db.Record):
+    def check_record(self, record: db.Record, identifiable: Identifiable):
         """
         Check for a record from the local storage (named "record") if it is
         the identified record for an identifiable which was created by
@@ -348,13 +361,11 @@ class LocalStorageIdentifiableAdapter(IdentifiableAdapter):
         record is the record from the local database to check against.
         identifiable is the record that was created during the crawler run.
         """
-        if len(identifiable.parents) != 1:
-            raise RuntimeError(
-                "Multiple parents for identifiables not supported.")
-        if not has_parent(record, identifiable.parents[0].name):
+        if (identifiable.record_type is not None
+                and not has_parent(record, identifiable.record_type)):
             return False
-        for prop in identifiable.properties:
-            prop_record = record.get_property(prop.name)
+        for propname, propvalue in identifiable.properties.items():
+            prop_record = record.get_property(propname)
             if prop_record is None:
                 return False
 
@@ -362,18 +373,18 @@ class LocalStorageIdentifiableAdapter(IdentifiableAdapter):
             # there are two different cases:
             # a) prop_record.value has a registered identifiable:
             #      in this case, fetch the identifiable and set the value accordingly
-            if isinstance(prop.value, db.Entity):  # lists are not checked here
+            if isinstance(propvalue, db.Entity):  # lists are not checked here
                 otherid = prop_record.value
                 if isinstance(prop_record.value, db.Entity):
                     otherid = prop_record.value.id
-                if prop.value.id != otherid:
+                if propvalue.id != otherid:
                     return False
 
-            elif prop.value != prop_record.value:
+            elif propvalue != prop_record.value:
                 return False
         return True
 
-    def retrieve_identified_record_for_identifiable(self, identifiable: db.Record):
+    def retrieve_identified_record_for_identifiable(self, identifiable: Identifiable):
         candidates = []
         for record in self._records:
             if self.check_record(record, identifiable):
@@ -420,13 +431,20 @@ class CaosDBIdentifiableAdapter(IdentifiableAdapter):
         for key, value in identifiable_data.items():
             rt = db.RecordType().add_parent(key)
             for prop_name in value:
-                rt.add_property(name=prop_name)
+                if isinstance(prop_name, str):
+                    rt.add_property(name=prop_name)
+                elif isinstance(prop_name, dict):
+                    for k, v in prop_name.items():
+                        rt.add_property(name=k, value=v)
+                else:
+                    NotImplementedError("YAML is not structured correctly")
+
             self.register_identifiable(key, rt)
 
     def register_identifiable(self, name: str, definition: db.RecordType):
         self._registered_identifiables[name] = definition
 
-    def get_file(self, identifiable: db.File):
+    def get_file(self, identifiable: Identifiable):
         if identifiable.path is None:
             raise RuntimeError("Path must not be None for File retrieval.")
         candidates = db.execute_query("FIND File which is stored at {}".format(
@@ -444,6 +462,9 @@ class CaosDBIdentifiableAdapter(IdentifiableAdapter):
         It is assumed, that there is exactly one identifiable for each RecordType. Only the first
         parent of the given Record is considered; others are ignored
         """
+        if len(record.parents) == 0:
+            return None
+        # TODO We need to treat the case where multiple parents exist properly.
         rt_name = record.parents[0].name
         for name, definition in self._registered_identifiables.items():
             if definition.parents[0].name.lower() == rt_name.lower():
@@ -458,12 +479,14 @@ class CaosDBIdentifiableAdapter(IdentifiableAdapter):
             return record
         return record.id
 
-    def retrieve_identified_record_for_identifiable(self, identifiable: db.Record):
+    def retrieve_identified_record_for_identifiable(self, identifiable: Identifiable):
         query_string = self.create_query_for_identifiable(identifiable)
         candidates = db.execute_query(query_string)
         if len(candidates) > 1:
             raise RuntimeError(
-                f"Identifiable was not defined unambigiously.\n{query_string}\nReturned the following {candidates}.")
+                f"Identifiable was not defined unambigiously.\n{query_string}\nReturned the "
+                f"following {candidates}."
+                f"Identifiable:\n{identifiable.record_type}{identifiable.properties}")
         if len(candidates) == 0:
             return None
         return candidates[0]
